/* ============================================================
 * File  : streamsniff.c
 * Author: Eric Giesselbach <ericgies@kabelfoon.nl>
 * Date  : 2005-09-01
 * Description : sniffs network for mms and rtsp stream url's 
 *
 * Copyright 2005 by Eric Giesselbach

 * This program is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published bythe Free Software Foundation;
 * either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * ============================================================ */

/*
  Todo: 
    OK if a stream url is detected, trace back (http) packets between 
       same server and client looking for (more stable) playlist url
    2. cache all http packets. Find stream url in non-GET http packet 
       data and trace back same connection (host, port) for GET.
    3. create separate protocol handlers with proper interface.
    4. Windows port?
    5. tdiff as command line parameter
*/ 

#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <linux/if_ether.h>
#include <linux/in.h>
#include "protocols/extern.h"
#include "protocols/mms.c"
#include "protocols/icy.c"
#include "protocols/rtsp.c"


FILE *dumpfile;  // ethereal saved packet for debugging
int sock;        // raw socket
int debug;       // to debug or not to debug
       
struct proto_ip
{
  char sep1[14]; // ethernet
  char sep2[9];
  char proto[1]; // 0x06 TCP, 0x11 UDP
  char sep3[2];
  struct in_addr saddr;
  struct in_addr daddr;
} __attribute__ ((packed));


struct proto_udp
{
  char sep[8];
} __attribute__ ((packed));


struct proto_tcp
{
  unsigned short sport; // use ntohs()
  unsigned short dport;
  char sep1[8];
  unsigned char offset[1];  // tcp header size = offset >> 4 * 4;
  char sep2[7];
  // + options, padding
} __attribute__ ((packed));

 

struct proto_http
{
   char command[5];
} __attribute__ ((packed));


void wipeall(int nogo)
{
  close(sock);
  if (debug == 1) fclose(dumpfile); 
  exit(1);  
}


enum parserstate
{
   searching = 0,
   streamfound,
   nostreamfound
};

int main(int argc, char *argv[])
{
       debug = 0;
 
       char *buffer, *curchar, *endchar;
       char buf[BUFLEN], host[200];
       int readcnt, i;
       size_t cursize;
       
       enum parserstate state;
            
       unsigned int test;
       
       struct connection _connection;
       struct connection *m_connection = &_connection;
     
       struct streamurl  *url;    
       struct proto_ip   *prot_ip;
       struct proto_udp  *prot_udp;       
       struct proto_tcp  *prot_tcp;
       struct proto_http *prot_http;

       buffer = buf;
       buffer[BUFLEN - 1] = 0;
       
       struct url *myUrl;
       
       int headerindex = 0;
       
       for (i=0; i < 100; i++) { 
        headers[i].header[0] = 0; 
        headers[i].headerlen = 0; 
       }
       
       signal(SIGINT,  wipeall);
           
       if ( (sock = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL))) < 0 ) 
       {
         printf("cannot open raw socket (you root?)\n");
         exit(-1);
       }
       
       if ( debug == 1 && (dumpfile = fopen("test.dump", "rb")) == NULL )
       {
          printf("Cannot open file test.dump\n");
          close(sock);
          exit(-1);
       }
            
       printf("looking for streams...\n\n");
       
       while (1) 
       {
         state = searching;
         
         if ( debug == 1 )
         {
           readcnt = fread(buf, 1, BUFLEN, dumpfile);
           buffer = buf + 40;
           if (readcnt >= 40) {readcnt -= 40;}
         }
           else
             readcnt = recvfrom(sock, buffer, BUFLEN - 1, 0, NULL, NULL);
        
         cursize = 0;

   // check min packet size   
         if (readcnt < (cursize + sizeof(struct proto_ip)) )
         {
           printf("incomplete header, aborting\n");
           close (sock);
           if (debug == 1) fclose(dumpfile);
           exit(-1);
         }
         
   // make sure char data terminates
          buffer[readcnt] = 0; 
  
   // map eth/ip structure (both assumed)
          prot_ip = (struct proto_ip*)buffer;        
          cursize += sizeof(struct proto_ip);  // should allow for options...
         
   // map tcp structure (prot_ip->proto == 0x06 assumed)
          if ( readcnt < cursize + sizeof(struct proto_tcp) ) 
            state = nostreamfound;
          else 
          {
            prot_tcp = (struct proto_tcp*)(buffer + cursize);
            cursize += (prot_tcp->offset[0] >> 4) * 4; // tcp header size
          }
          
   // register connection endpoints
          if (state == searching)
          {
             strcpy(m_connection->saddr, (char*)inet_ntoa(prot_ip->saddr));
             strcpy(m_connection->daddr, (char*)inet_ntoa(prot_ip->daddr));
             m_connection->sport = prot_tcp->sport;            
             m_connection->dport = prot_tcp->dport;            
          } 
          
   // check http
         if ( state == searching )
            if (readcnt < cursize + sizeof(struct proto_http) )
               state = nostreamfound;
            else
            {
               prot_http = (struct proto_http*)(buffer + cursize);
            
               // store http headers
               if ( strncmp(prot_http->command, "GET " , 4) == 0 ||
                    strncmp(prot_http->command, "HTTP/", 5) == 0   )
               {
                  if (++headerindex > 99) headerindex = 0;
               
                  strcpy(headers[headerindex].conn.saddr, (char*)inet_ntoa(prot_ip->saddr));
                  strcpy(headers[headerindex].conn.daddr, (char*)inet_ntoa(prot_ip->daddr));
               
                  headers[headerindex].conn.sport = prot_tcp->sport;
                  headers[headerindex].conn.dport = prot_tcp->dport;
               
                  headers[headerindex].timestamp = time(0);
                  strncpy(headers[headerindex].header, prot_http->command, readcnt - cursize); 
                  headers[headerindex].headerlen = readcnt - cursize;
               }
             }

    // check rtsp
        if (state == searching)
        {
           
           url = check_rtsp(buffer + cursize, readcnt - cursize, headerindex);                
           
           if ( url->type == none ) 
             url = check_icy(buffer + cursize, readcnt - cursize, headerindex, m_connection); 
               
           if ( url->type == none ) 
             url = check_mms(buffer + cursize, readcnt - cursize, headerindex, m_connection); 
        }
        
           
           
         if ( url->type != none) 
         {
            printf ("%s: %s%s%s\n", names[url->type], url->prot, url->host, url->path);
            printf ("    : %s\n\n", url->plst);
            url->type = none;
         }
       
       }
             
} 
